@@ -28,7 +28,7 @@
 
  */
 
-/* $Id: apc_cache.c,v 3.122 2006/08/22 14:36:37 gopalv Exp $ */
+/* $Id: apc_cache.c,v 3.140 2007/04/02 22:57:10 rasmus Exp $ */
 
 #include "apc_cache.h"
 #include "apc_lock.h"
@@ -36,17 +36,18 @@
 #include "apc_globals.h"
 #include "SAPI.h"
 #include "ext/standard/php_var.h"
+#include "ext/standard/php_smart_str.h"
 
 /* TODO: rehash when load factor exceeds threshold */
 
 #define CHECK(p) { if ((p) == NULL) return NULL; }
 
 /* {{{ locking macros */
-#define CREATE_LOCK     apc_lck_create(NULL, 0, 1)
-#define DESTROY_LOCK(c) apc_lck_destroy(c->lock)
-#define LOCK(c)         { HANDLE_BLOCK_INTERRUPTIONS(); apc_lck_lock(c->lock); }
-#define RDLOCK(c)       { HANDLE_BLOCK_INTERRUPTIONS(); apc_lck_rdlock(c->lock); }
-#define UNLOCK(c)       { apc_lck_unlock(c->lock); HANDLE_UNBLOCK_INTERRUPTIONS(); }
+#define CREATE_LOCK(lock)     apc_lck_create(NULL, 0, 1, lock)
+#define DESTROY_LOCK(c) apc_lck_destroy(c->header->lock)
+#define LOCK(c)         { HANDLE_BLOCK_INTERRUPTIONS(); apc_lck_lock(c->header->lock); }
+#define RDLOCK(c)       { HANDLE_BLOCK_INTERRUPTIONS(); apc_lck_rdlock(c->header->lock); }
+#define UNLOCK(c)       { apc_lck_unlock(c->header->lock); HANDLE_UNBLOCK_INTERRUPTIONS(); }
 /* }}} */
 
 /* {{{ struct definition: slot_t */
@@ -66,8 +67,11 @@ struct slot_t {
    Any values that must be shared among processes should go in here. */
 typedef struct header_t header_t;
 struct header_t {
+    apc_lck_t lock;              /* read/write lock (exclusive blocking cache lock) */
+    apc_lck_t wrlock;           /* write lock (non-blocking used to prevent cache slams) */
     int num_hits;               /* total successful hits in cache */
     int num_misses;             /* total unsuccessful hits in cache */
+    int num_inserts;            /* total successful inserts in cache */
     slot_t* deleted_list;       /* linked list of to-be-deleted slots */
     time_t start_time;          /* time the above counters were reset */
     int expunges;               /* total number of expunges */
@@ -85,8 +89,27 @@ struct apc_cache_t {
     int num_slots;              /* number of slots in cache */
     int gc_ttl;                 /* maximum time on GC list for a slot */
     int ttl;                    /* if slot is needed and entry's access time is older than this ttl, remove it */
-    int lock;                   /* read/write lock (exclusive blocking cache lock) */
-    int wrlock;                 /* write lock (non-blocking used to prevent cache slams) */
+};
+/* }}} */
+
+/* {{{ struct definition local_slot_t */
+typedef struct local_slot_t local_slot_t;
+struct local_slot_t {
+    slot_t *original;           /* the original slot in shm */
+    int num_hits;               /* number of hits */
+    apc_cache_entry_t *value;   /* shallow copy of slot->value */
+    local_slot_t *next;         /* only for dead list */
+};
+/* }}} */
+/* {{{ struct definition apc_local_cache_t */
+struct apc_local_cache_t {
+    apc_cache_t* shmcache;      /* the real cache in shm */
+    local_slot_t* slots;        /* process (local) cache of objects */
+    local_slot_t* dead_list;    /* list of objects pending removal */
+    int num_slots;              /* number of slots in cache */
+    int ttl;                    /* time to live */
+    int num_hits;               /* number of hits */
+    int generation;             /* every generation lives between expunges */
 };
 /* }}} */
 
@@ -297,9 +320,9 @@ apc_cache_t* apc_cache_create(int size_hint, int gc_ttl, int ttl)
     cache->num_slots = num_slots;
     cache->gc_ttl = gc_ttl;
     cache->ttl = ttl;
-    cache->lock   = CREATE_LOCK;
+    CREATE_LOCK(cache->header->lock);
 #if NONBLOCKING_LOCK_AVAILABLE
-    cache->wrlock = CREATE_LOCK;
+    CREATE_LOCK(cache->header->wrlock);
 #endif
     for (i = 0; i < num_slots; i++) {
         cache->slots[i] = NULL;
@@ -473,6 +496,7 @@ int apc_cache_insert(apc_cache_t* cache,
    
     cache->header->mem_size += value->mem_size;
     cache->header->num_entries++;
+    cache->header->num_inserts++;
     
     UNLOCK(cache);
     return 1;
@@ -480,7 +504,7 @@ int apc_cache_insert(apc_cache_t* cache,
 /* }}} */
 
 /* {{{ apc_cache_user_insert */
-int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_entry_t* value, time_t t TSRMLS_DC)
+int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_entry_t* value, time_t t, int exclusive TSRMLS_DC)
 {
     slot_t** slot;
     size_t* mem_size_ptr = NULL;
@@ -501,7 +525,18 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
 
     while (*slot) {
         if (!memcmp((*slot)->key.data.user.identifier, key.data.user.identifier, key.data.user.identifier_len)) {
-            /* If a slot with the same identifier already exists, remove it */
+            /* 
+             * At this point we have found the user cache entry.  If we are doing 
+             * an exclusive insert (apc_add) we are going to bail right away if
+             * the user entry already exists and it has no ttl, or
+             * there is a ttl and the entry has not timed out yet.
+             */
+            if(exclusive && (  !(*slot)->value->data.user.ttl ||
+                              ( (*slot)->value->data.user.ttl && ((*slot)->creation_time + (*slot)->value->data.user.ttl) >= t ) 
+                            ) ) {
+                UNLOCK(cache);
+                return 0;
+            }
             remove_slot(cache, slot);
             break;
         } else 
@@ -533,16 +568,18 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
         cache->header->mem_size += *APCG(mem_size_ptr);
     }
     cache->header->num_entries++;
+    cache->header->num_inserts++;
 
     UNLOCK(cache);
     return 1;
 }
 /* }}} */
 
-/* {{{ apc_cache_find */
-apc_cache_entry_t* apc_cache_find(apc_cache_t* cache, apc_cache_key_t key, time_t t)
+/* {{{ apc_cache_find_slot */
+slot_t* apc_cache_find_slot(apc_cache_t* cache, apc_cache_key_t key, time_t t)
 {
     slot_t** slot;
+    volatile slot_t* retval = NULL;
 
     LOCK(cache);
     if(key.type == APC_CACHE_KEY_FILE) slot = &cache->slots[hash(key) % cache->num_slots];
@@ -563,11 +600,9 @@ apc_cache_entry_t* apc_cache_find(apc_cache_t* cache, apc_cache_key_t key, time_
                 (*slot)->access_time = t;
                 prevent_garbage_collection((*slot)->value);
                 cache->header->num_hits++;
-                if(1) {
-                    apc_cache_entry_t* value = (*slot)->value;
-                    UNLOCK(cache);
-                    return value;
-                }
+                retval = *slot;
+                UNLOCK(cache);
+                return (slot_t*)retval;
             }
         } else {  /* APC_CACHE_KEY_FPFILE */
             if(!memcmp((*slot)->key.data.fpfile.fullpath, key.data.fpfile.fullpath, key.data.fpfile.fullpath_len+1)) {
@@ -577,11 +612,9 @@ apc_cache_entry_t* apc_cache_find(apc_cache_t* cache, apc_cache_key_t key, time_
                 (*slot)->access_time = t;
                 prevent_garbage_collection((*slot)->value);
                 cache->header->num_hits++;
-                if(1) {
-                    apc_cache_entry_t* value = (*slot)->value;
-                    UNLOCK(cache);
-                    return value;
-                }
+                retval = *slot;
+                UNLOCK(cache);
+                return (slot_t*)retval;
             }
         }
       }
@@ -593,10 +626,19 @@ apc_cache_entry_t* apc_cache_find(apc_cache_t* cache, apc_cache_key_t key, time_
 }
 /* }}} */
 
+/* {{{ apc_cache_find */
+apc_cache_entry_t* apc_cache_find(apc_cache_t* cache, apc_cache_key_t key, time_t t)
+{
+    slot_t * slot = apc_cache_find_slot(cache, key, t);
+    return (slot) ? slot->value : NULL;
+}
+/* }}} */
+
 /* {{{ apc_cache_user_find */
 apc_cache_entry_t* apc_cache_user_find(apc_cache_t* cache, char *strkey, int keylen, time_t t)
 {
     slot_t** slot;
+    volatile apc_cache_entry_t* value = NULL;
 
     LOCK(cache);
 
@@ -616,11 +658,9 @@ apc_cache_entry_t* apc_cache_user_find(apc_cache_t* cache, char *strkey, int key
             (*slot)->access_time = t;
 
             cache->header->num_hits++;
-            if(1) {
-                apc_cache_entry_t * value = (*slot)->value;
-                UNLOCK(cache);
-                return value;
-            }
+            value = (*slot)->value;
+            UNLOCK(cache);
+            return (apc_cache_entry_t*)value;
         }
         slot = &(*slot)->next;
     }
@@ -656,6 +696,9 @@ int apc_cache_user_delete(apc_cache_t* cache, char *strkey, int keylen)
 /* {{{ apc_cache_release */
 void apc_cache_release(apc_cache_t* cache, apc_cache_entry_t* entry)
 {
+    /* local cache refcount-- is done in apc_local_cache_cleanup */
+    if(entry->local) return;
+
     LOCK(cache);
     entry->ref_count--;
     UNLOCK(cache);
@@ -669,8 +712,9 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
                        time_t t
 					   TSRMLS_DC)
 {
-    static char canon_path[PATH_MAX];
-    struct stat buf, *tmp_buf=NULL;
+    static char canon_path[MAXPATHLEN];
+    struct stat *tmp_buf=NULL;
+    struct apc_fileinfo_t fileinfo = { {0}, };
     int len;
 	
     assert(key != NULL);
@@ -706,9 +750,9 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
         tmp_buf = sapi_get_stat(TSRMLS_C);  /* Apache has already done this stat() for us */
     }
     if(tmp_buf) { 
-		buf = *tmp_buf;
+		fileinfo.st_buf = *tmp_buf;
     } else {
-        if (apc_stat_paths(filename, include_path, &buf) != 0) {
+        if (apc_search_paths(filename, include_path, &fileinfo) != 0) {
 #ifdef __DEBUG_APC__
             fprintf(stderr,"Stat failed %s - bailing (%s) (%d)\n",filename,SG(request_info).path_translated);
 #endif
@@ -716,9 +760,9 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
         }
     }
 
-    if(APCG(max_file_size) < buf.st_size) {
+    if(APCG(max_file_size) < fileinfo.st_buf.st_size) {
 #ifdef __DEBUG_APC__
-        fprintf(stderr,"File is too big %s (%d - %ld) - bailing\n",filename,t,buf.st_size);
+        fprintf(stderr,"File is too big %s (%d - %ld) - bailing\n",filename,t,fileinfo.st_buf.st_size);
 #endif
         return 0;
     }
@@ -735,15 +779,15 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
      * tiny safety is easier than educating the world.  This is now
      * configurable, but the default is still 2 seconds.
      */
-    if(APCG(file_update_protection) && (t - buf.st_mtime < APCG(file_update_protection))) { 
+    if(APCG(file_update_protection) && (t - fileinfo.st_buf.st_mtime < APCG(file_update_protection))) { 
 #ifdef __DEBUG_APC__
-        fprintf(stderr,"File is too new %s (%d - %d) - bailing\n",filename,t,buf.st_mtime);
+        fprintf(stderr,"File is too new %s (%d - %d) - bailing\n",filename,t,fileinfo.st_buf.st_mtime);
 #endif
         return 0;
     }
 
-    key->data.file.device = buf.st_dev;
-    key->data.file.inode  = buf.st_ino;
+    key->data.file.device = fileinfo.st_buf.st_dev;
+    key->data.file.inode  = fileinfo.st_buf.st_ino;
     /* 
      * If working with content management systems that like to munge the mtime, 
      * it might be appropriate to key off of the ctime to be immune to systems
@@ -753,11 +797,13 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
      * I generally disagree with using the ctime here because you lose the 
      * ability to warm up new content by saving it to a temporary file, hitting
      * it once to cache it and then renaming it into its permanent location so
-     * I am leaving this commented out by default.  If you need this, uncomment
-     * it.
+     * set the apc.stat_ctime=true to enable this check.
      */
-    /* key->mtime  = (buf.st_ctime > buf.st_mtime) ? buf.st_ctime : buf.st_mtime; */
-    key->mtime = buf.st_mtime;
+    if(APCG(stat_ctime)) {
+        key->mtime  = (fileinfo.st_buf.st_ctime > fileinfo.st_buf.st_mtime) ? fileinfo.st_buf.st_ctime : fileinfo.st_buf.st_mtime; 
+    } else {
+        key->mtime = fileinfo.st_buf.st_mtime;
+    }
     key->type = APC_CACHE_KEY_FILE;
     return 1;
 }
@@ -807,6 +853,8 @@ apc_cache_entry_t* apc_cache_make_file_entry(const char* filename,
     entry->type = APC_CACHE_ENTRY_FILE;
     entry->ref_count = 0;
     entry->mem_size = 0;
+    entry->autofiltered = 0;
+    entry->local = 0;
     return entry;
 }
 /* }}} */
@@ -816,10 +864,9 @@ zval* apc_cache_store_zval(zval* dst, const zval* src, apc_malloc_t allocate, ap
 {
     smart_str buf = {0};
     php_serialize_data_t var_hash;
+    TSRMLS_FETCH();
 
     if((src->type & ~IS_CONSTANT_INDEX) == IS_OBJECT) {
-        TSRMLS_FETCH();
-
         if(!dst) {
             CHECK(dst = (zval*) allocate(sizeof(zval)));
         }
@@ -834,10 +881,26 @@ zval* apc_cache_store_zval(zval* dst, const zval* src, apc_malloc_t allocate, ap
             dst->value.str.len = buf.len;
             CHECK(dst->value.str.val = apc_xmemcpy(buf.c, buf.len+1, allocate));
             dst->type = src->type;
+            smart_str_free(&buf);
         }
         return dst; 
     } else {
-        return apc_copy_zval(dst, src, allocate, deallocate);
+        
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        HashTable *old = APCG(copied_zvals);
+        APCG(copied_zvals) = emalloc(sizeof(HashTable));
+        zend_hash_init(APCG(copied_zvals), 0, NULL, NULL, 0);
+        
+        dst = apc_copy_zval(dst, src, allocate, deallocate);
+
+        if(APCG(copied_zvals)) {
+            zend_hash_destroy(APCG(copied_zvals));
+            efree(APCG(copied_zvals));
+        }
+
+        APCG(copied_zvals) = old;
+
+        return dst;
     }
 }
 /* }}} */
@@ -845,10 +908,10 @@ zval* apc_cache_store_zval(zval* dst, const zval* src, apc_malloc_t allocate, ap
 /* {{{ apc_cache_fetch_zval */
 zval* apc_cache_fetch_zval(zval* dst, const zval* src, apc_malloc_t allocate, apc_free_t deallocate)
 {
+    TSRMLS_FETCH();
     if((src->type & ~IS_CONSTANT_INDEX) == IS_OBJECT) {
         php_unserialize_data_t var_hash;
         const unsigned char *p = (unsigned char*)Z_STRVAL_P(src);
-        TSRMLS_FETCH();
 
         PHP_VAR_UNSERIALIZE_INIT(var_hash);
         if(!php_var_unserialize(&dst, &p, p + Z_STRLEN_P(src), &var_hash TSRMLS_CC)) {
@@ -860,7 +923,22 @@ zval* apc_cache_fetch_zval(zval* dst, const zval* src, apc_malloc_t allocate, ap
         PHP_VAR_UNSERIALIZE_DESTROY(var_hash);		
         return dst; 
     } else {
-        return apc_copy_zval(dst, src, allocate, deallocate);
+    
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        HashTable *old = APCG(copied_zvals);
+        APCG(copied_zvals) = emalloc(sizeof(HashTable));
+        zend_hash_init(APCG(copied_zvals), 0, NULL, NULL, 0);
+        
+        dst = apc_copy_zval(dst, src, allocate, deallocate);
+
+        if(APCG(copied_zvals)) {
+            zend_hash_destroy(APCG(copied_zvals));
+            efree(APCG(copied_zvals));
+        }
+
+        APCG(copied_zvals) = old;
+
+        return dst;
     }
 }
 /* }}} */
@@ -868,13 +946,26 @@ zval* apc_cache_fetch_zval(zval* dst, const zval* src, apc_malloc_t allocate, ap
 /* {{{ apc_cache_free_zval */
 void apc_cache_free_zval(zval* src, apc_free_t deallocate)
 {
+    TSRMLS_FETCH();
     if ((src->type & ~IS_CONSTANT_INDEX) == IS_OBJECT) {
         if (src->value.str.val) {
         	deallocate(src->value.str.val);
         }
         deallocate(src);
     } else {
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        HashTable *old = APCG(copied_zvals);
+        APCG(copied_zvals) = emalloc(sizeof(HashTable));
+        zend_hash_init(APCG(copied_zvals), 0, NULL, NULL, 0);
+        
         apc_free_zval(src, deallocate);
+
+        if(APCG(copied_zvals)) {
+            zend_hash_destroy(APCG(copied_zvals));
+            efree(APCG(copied_zvals));
+        }
+
+        APCG(copied_zvals) = old;
     }
 }
 /* }}} */
@@ -904,6 +995,8 @@ apc_cache_entry_t* apc_cache_make_user_entry(const char* info, int info_len, con
     entry->type = APC_CACHE_ENTRY_USER;
     entry->ref_count = 0;
     entry->mem_size = 0;
+    entry->autofiltered = 0;
+    entry->local = 0;
     return entry;
 }
 /* }}} */
@@ -956,6 +1049,7 @@ apc_cache_info_t* apc_cache_info(apc_cache_t* cache, zend_bool limited)
     info->expunges = cache->header->expunges;
     info->mem_size = cache->header->mem_size;
     info->num_entries = cache->header->num_entries;
+    info->num_inserts = cache->header->num_inserts;
 
     if(!limited) {
         /* For each hashtable slot */
@@ -1064,18 +1158,166 @@ zend_bool apc_cache_busy(apc_cache_t* cache)
 /* {{{ apc_cache_write_lock */
 zend_bool apc_cache_write_lock(apc_cache_t* cache)
 {
-    return apc_lck_nb_lock(cache->wrlock);
+    return apc_lck_nb_lock(cache->header->wrlock);
 }
 /* }}} */
 
 /* {{{ apc_cache_write_unlock */
 void apc_cache_write_unlock(apc_cache_t* cache)
 {
-    apc_lck_unlock(cache->wrlock);
+    apc_lck_unlock(cache->header->wrlock);
 }
 /* }}} */
 #endif
 
+/* {{{ make_local_slot */
+static local_slot_t* make_local_slot(apc_local_cache_t* cache, local_slot_t* lslot, slot_t* slot) 
+{
+    apc_cache_entry_t* value;
+
+    value = apc_emalloc(sizeof(apc_cache_entry_t));
+    memcpy(value, slot->value, sizeof(apc_cache_entry_t)); /* bitwise copy */
+    value->local = 1;
+
+    lslot->original = slot;
+    lslot->value = value;
+    lslot->num_hits++;
+
+    return lslot; /* for what joy ? ... consistency */
+}
+/* }}} */
+
+/* {{{ free_local_slot */
+static void free_local_slot(apc_local_cache_t* cache, local_slot_t* lslot) 
+{
+    local_slot_t * dead = NULL;
+    if(!lslot->original) return;
+
+    /* TODO: Bad design to allocate memory in a free_* - fix when bored (hehe) */
+    dead = apc_emalloc(sizeof(local_slot_t));
+    memcpy(dead, lslot, sizeof(local_slot_t)); /* bitwise copy */
+
+    lslot->original = NULL;
+    lslot->value = NULL;
+
+    dead->next = cache->dead_list;
+    cache->dead_list = dead;
+}
+/* }}} */
+
+/* {{{ apc_local_cache_create */
+apc_local_cache_t* apc_local_cache_create(apc_cache_t *shmcache, int num_slots, int ttl)
+{
+    apc_local_cache_t* cache = NULL;
+
+    cache = (apc_local_cache_t*) apc_emalloc(sizeof(apc_local_cache_t));
+
+    cache->slots = (local_slot_t*) (apc_emalloc(sizeof(local_slot_t) * num_slots));
+    memset(cache->slots, 0, sizeof(local_slot_t) * num_slots);
+
+    cache->shmcache = shmcache;
+    cache->num_slots = num_slots;
+    cache->ttl = ttl;
+    cache->num_hits = 0;
+    cache->generation = shmcache->header->expunges;
+    cache->dead_list = NULL;
+
+    return cache;
+}
+/* }}} */
+
+/* {{{ apc_local_cache_cleanup */
+void apc_local_cache_cleanup(apc_local_cache_t* cache) {
+    local_slot_t * lslot;
+    time_t t = time(0);
+    
+    int i;
+    for(i = 0; i < cache->num_slots; i++) {
+        slot_t * slot = cache->slots[i].original;
+        if((slot && slot->access_time < (t - cache->ttl)) ||
+                cache->generation != cache->shmcache->header->expunges) {
+            free_local_slot(cache, &cache->slots[i]);
+        }
+    }
+
+    LOCK(cache->shmcache);
+    for(lslot = cache->dead_list; lslot != NULL; lslot = lslot->next) {
+        lslot->original->num_hits += lslot->num_hits;
+        lslot->original->value->ref_count--; /* apc_cache_release(cache->shmcache, lslot->original->value); */
+        apc_efree(lslot->value);
+    }
+    UNLOCK(cache->shmcache);
+
+    cache->dead_list = NULL;
+}
+/* }}} */
+
+/* {{{ apc_local_cache_destroy */
+void apc_local_cache_destroy(apc_local_cache_t* cache)
+{
+    int i;
+    for(i = 0; i < cache->num_slots; i++) {
+        free_local_slot(cache, &cache->slots[i]);
+    }
+
+    apc_local_cache_cleanup(cache);
+
+    LOCK(cache->shmcache);
+    cache->shmcache->header->num_hits += cache->num_hits;
+    UNLOCK(cache->shmcache);
+
+    apc_efree(cache->slots);
+    apc_efree(cache);
+}
+/* }}} */
+
+/* {{{ apc_local_cache_find */
+apc_cache_entry_t* apc_local_cache_find(apc_local_cache_t* cache, apc_cache_key_t key, time_t t)
+{
+    slot_t* slot;
+    local_slot_t* lslot; 
+
+    if(key.type == APC_CACHE_KEY_FILE) lslot = &cache->slots[hash(key) % cache->num_slots];
+    else lslot = &cache->slots[string_nhash_8(key.data.fpfile.fullpath, key.data.fpfile.fullpath_len) % cache->num_slots];
+
+    slot = lslot->original;
+
+    if(slot && key.type == slot->key.type) {
+        if(slot->access_time < (t - cache->ttl)) {
+            goto not_found;
+        }
+        if(key.type == APC_CACHE_KEY_FILE && 
+           key_equals(slot->key.data.file, key.data.file)) {
+            if(slot->key.mtime != key.mtime) {
+                free_local_slot(cache, lslot);
+                goto not_found;
+            }
+            return lslot->value;
+        } else if(key.type == APC_CACHE_KEY_FPFILE) {
+            if(!memcmp(slot->key.data.fpfile.fullpath, key.data.fpfile.fullpath, key.data.fpfile.fullpath_len+1)) {
+                return lslot->value;
+            }
+        }
+    }
+not_found:
+    if(apc_cache_busy(cache->shmcache)) {
+        return NULL;
+    }
+
+    slot = apc_cache_find_slot(cache->shmcache, key, t);
+
+    if(!slot) return NULL;
+   
+    /* i.e maintain a sort of top list */
+    if(lslot->original == NULL || (lslot->original->num_hits + lslot->num_hits)  < slot->num_hits) {
+        free_local_slot(cache, lslot);
+        make_local_slot(cache, lslot, slot); 
+        return lslot->value;
+    }
+    return slot->value;
+}
+/* }}} */
+
 /*
  * Local variables:
  * tab-width: 4
