@@ -28,12 +28,11 @@
 
  */
 
-/* $Id: apc_compile.c,v 3.58 2006/09/05 06:31:56 gopalv Exp $ */
+/* $Id: apc_compile.c,v 3.85 2007/03/28 07:35:55 gopalv Exp $ */
 
 #include "apc_compile.h"
 #include "apc_globals.h"
 #include "apc_zend.h"
-#include "apc_optimizer.h"
 
 typedef void* (*ht_copy_fun_t)(void*, void*, apc_malloc_t, apc_free_t);
 typedef void  (*ht_free_fun_t)(void*, apc_free_t);
@@ -75,9 +74,12 @@ static zend_arg_info* my_copy_arg_info(zend_arg_info*, zend_arg_info*, apc_mallo
 /*
  * The "destroy" functions free the memory associated with a particular data
  * structure but do not free the pointer to the data structure.
+ *
+ * my_destroy_zval() returns SUCCESS or FAILURE, FAILURE means that
+ * the zval* has other references elsewhere 
  */
+static int  my_destroy_zval(zval*, apc_free_t); 
 static void my_destroy_zval_ptr(zval**, apc_free_t);
-static void my_destroy_zval(zval*, apc_free_t);
 static void my_destroy_zend_op(zend_op*, apc_free_t);
 static void my_destroy_znode(znode*, apc_free_t);
 static void my_destroy_function(zend_function*, apc_free_t);
@@ -241,7 +243,8 @@ static zend_function* my_bitwise_copy_function(zend_function* dst, zend_function
 static zval** my_copy_zval_ptr(zval** dst, const zval** src, apc_malloc_t allocate, apc_free_t deallocate)
 {
     int local_dst_alloc = 0;
-
+    zval* dst_new;
+    
     assert(src != NULL);
 
     if (!dst) {
@@ -253,11 +256,14 @@ static zval** my_copy_zval_ptr(zval** dst, const zval** src, apc_malloc_t alloca
         if(local_dst_alloc) deallocate(dst);
         return NULL;
     }
-    if(!my_copy_zval(*dst, *src, allocate, deallocate)) return NULL;
+    dst_new = my_copy_zval(*dst, *src, allocate, deallocate);
+    if(dst_new != *dst) {
+        deallocate(*dst);
+        *dst = dst_new;
+    }
 
-    /* deep-copying ensures that there is only one reference to this in memory */
-    (*dst)->refcount = 1;
-    (*dst)->is_ref = 0;
+    (*dst)->refcount = (*src)->refcount;
+    (*dst)->is_ref = (*src)->is_ref;
     
     return dst;
 }
@@ -266,6 +272,9 @@ static zval** my_copy_zval_ptr(zval** dst, const zval** src, apc_malloc_t alloca
 /* {{{ my_copy_zval */
 static zval* my_copy_zval(zval* dst, const zval* src, apc_malloc_t allocate, apc_free_t deallocate)
 {
+    zval **tmp;
+    TSRMLS_FETCH();
+    
     assert(dst != NULL);
     assert(src != NULL);
 
@@ -292,7 +301,19 @@ static zval* my_copy_zval(zval* dst, const zval* src, apc_malloc_t allocate, apc
         break;
     
     case IS_ARRAY:
+
+        if(APCG(copied_zvals)) {
+            if(zend_hash_index_find(APCG(copied_zvals), (ulong)src, (void**)&tmp) == SUCCESS) {
+                (*tmp)->refcount++;
+                return *tmp;
+            }
+        
+            zend_hash_index_update(APCG(copied_zvals), (ulong)src, (void**)&dst, sizeof(zval*), NULL);
+        }
+        /* fall through */
+ 
     case IS_CONSTANT_ARRAY:
+
         CHECK(dst->value.ht =
             my_copy_hashtable(NULL,
                               src->value.ht,
@@ -764,7 +785,8 @@ static zend_class_entry* my_copy_class_entry(zend_class_entry* dst, zend_class_e
                             1,
                             allocate, deallocate,
                             (ht_check_copy_fun_t) my_check_copy_static_member,
-                            src)) {
+                            src,
+                            &src->default_static_members)) {
         goto cleanup;
     }
     if(src->static_members != &src->default_static_members)
@@ -776,7 +798,8 @@ static zend_class_entry* my_copy_class_entry(zend_class_entry* dst, zend_class_e
                             1,
                             allocate, deallocate,
                             (ht_check_copy_fun_t) my_check_copy_static_member,
-                            src))) {
+                            src,
+                            src->static_members))) {
             goto cleanup;
         }
     }
@@ -855,7 +878,7 @@ cleanup:
     if(dst->properties_info.arBuckets) my_destroy_hashtable(&dst->properties_info, (ht_free_fun_t) my_free_property_info, deallocate);
     if(dst->default_static_members.arBuckets)
     {
-        my_destroy_hashtable(dst->static_members, (ht_free_fun_t) my_free_zval_ptr, deallocate);
+        my_destroy_hashtable(&dst->default_static_members, (ht_free_fun_t) my_free_zval_ptr, deallocate);
     }
     if(dst->static_members && dst->static_members != &(dst->default_static_members))
     {
@@ -1027,7 +1050,8 @@ zval* apc_copy_zval(zval* dst, const zval* src, apc_malloc_t allocate, apc_free_
         local_dst_alloc = 1;
     }
 
-    if(!my_copy_zval(dst, src, allocate, deallocate)) {
+    dst = my_copy_zval(dst, src, allocate, deallocate);
+    if(!dst) {
         if(local_dst_alloc) deallocate(dst);
         return NULL;
     }
@@ -1068,6 +1092,9 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
 {
     int i;
     int local_dst_alloc = 0;
+    apc_fileinfo_t fileinfo;
+    char canon_path[MAXPATHLEN];
+    char *fullpath = NULL;
 #ifdef ZEND_ENGINE_2
     apc_opflags_t * flags = NULL;
 #endif
@@ -1078,10 +1105,11 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
         CHECK(dst = (zend_op_array*) allocate(sizeof(src[0])));
         local_dst_alloc = 1;
     }
-    if(APCG(optimization)) {
-        apc_optimize_op_array(src TSRMLS_CC);
-    }
 
+    if(APCG(apc_optimize_function)) {
+        APCG(apc_optimize_function)(src TSRMLS_CC);
+    }
+    
     /* start with a bitwise copy of the array */
     memcpy(dst, src, sizeof(src[0]));
 
@@ -1146,9 +1174,13 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
     }
 
 #ifdef ZEND_ENGINE_2
-    flags = (apc_opflags_t*) & (dst->reserved);
-    memset(flags, 0, sizeof(apc_opflags_t));
-    /* assert(sizeof(apc_opflags_t) < sizeof(dst->reserved)); */
+    if(APCG(reserved_offset) != -1) {
+        /* Insanity alert: the void* pointer is cast into an apc_opflags_t 
+         * struct. apc_zend_init() checks to ensure that it fits in a void* */
+        flags = (apc_opflags_t*) & (dst->reserved[APCG(reserved_offset)]);
+        memset(flags, 0, sizeof(apc_opflags_t));
+        /* assert(sizeof(apc_opflags_t) < sizeof(dst->reserved)); */
+    }
 #endif
     
     for (i = 0; i < src->last; i++) {
@@ -1161,13 +1193,17 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
             case ZEND_JMPNZ:
             case ZEND_JMPZ_EX:
             case ZEND_JMPNZ_EX:
-                flags->has_jumps = 1;
+                if(flags != NULL) {
+                    flags->has_jumps = 1;
+                }
                 break;
 #ifdef ZEND_ENGINE_2
             /* auto_globals_jit was not in php-4.3.* */
             case ZEND_FETCH_R:
             case ZEND_FETCH_W:
-                if(PG(auto_globals_jit))
+            case ZEND_FETCH_IS:
+            case ZEND_FETCH_FUNC_ARG:
+                if(PG(auto_globals_jit) && flags != NULL)
                 {
                      /* The fetch is only required if auto_globals_jit=1  */
                     if(zo->op2.u.EA.type == ZEND_FETCH_GLOBAL &&
@@ -1175,7 +1211,23 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
                             zo->op1.u.constant.type == IS_STRING) {
                         znode * varname = &zo->op1;
                         if (varname->u.constant.value.str.val[0] == '_') {
-                            flags->use_globals = 1;
+#define SET_IF_AUTOGLOBAL(member) \
+    if(!strcmp(varname->u.constant.value.str.val, #member)) \
+        flags->member = 1 /* no ';' here */
+                            SET_IF_AUTOGLOBAL(_GET);
+                            else SET_IF_AUTOGLOBAL(_POST);
+                            else SET_IF_AUTOGLOBAL(_COOKIE);
+                            else SET_IF_AUTOGLOBAL(_SERVER);
+                            else SET_IF_AUTOGLOBAL(_ENV);
+                            else SET_IF_AUTOGLOBAL(_FILES);
+                            else SET_IF_AUTOGLOBAL(_REQUEST);
+                            else if(zend_is_auto_global(
+                                            varname->u.constant.value.str.val,
+                                            varname->u.constant.value.str.len
+                                            TSRMLS_CC))
+                            {
+                                flags->unknown_global = 1;
+                            }
                         }
                     }
                 }
@@ -1184,7 +1236,9 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
             case ZEND_RECV_INIT:
                 if(zo->op2.op_type == IS_CONST &&
                     zo->op2.u.constant.type == IS_CONSTANT_ARRAY) {
-                    flags->deep_copy = 1;
+                    if(flags != NULL) {
+                        flags->deep_copy = 1;
+                    }
                 }
                 break;
             default:
@@ -1192,7 +1246,9 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
                     zo->op1.u.constant.type == IS_CONSTANT_ARRAY) ||
                     (zo->op2.op_type == IS_CONST &&
                         zo->op2.u.constant.type == IS_CONSTANT_ARRAY)) {
-                    flags->deep_copy = 1;
+                    if(flags != NULL) {
+                        flags->deep_copy = 1;
+                    }
                 }
                 break;
         }
@@ -1204,10 +1260,31 @@ zend_op_array* apc_copy_op_array(zend_op_array* dst, zend_op_array* src, apc_mal
             }
             goto  cleanup;
         }
+#ifdef ZEND_ENGINE_2
+/* This code breaks apc's rule#1 - cache what you compile */
+        if(APCG(fpstat)==0) {
+            if((zo->opcode == ZEND_INCLUDE_OR_EVAL) && 
+                (zo->op1.op_type == IS_CONST && zo->op1.u.constant.type == IS_STRING)) {
+                /* constant includes */
+                if(!IS_ABSOLUTE_PATH(Z_STRVAL_P(&zo->op1.u.constant),len)) { 
+                    if (apc_search_paths(Z_STRVAL_P(&zo->op1.u.constant), PG(include_path), &fileinfo) == 0) {
+                        if((IS_ABSOLUTE_PATH(fileinfo.fullpath, strlen(fileinfo.fullpath)) && (fullpath = fileinfo.fullpath))
+                                || (fullpath = realpath(fileinfo.fullpath, canon_path))) {
+                            /* is either an absolute path or it goes through a realpath() */
+                            zend_op *dzo = &(dst->opcodes[i]);
+                            deallocate(dzo->op1.u.constant.value.str.val);
+                            dzo->op1.u.constant.value.str.len = strlen(fullpath);
+                            dzo->op1.u.constant.value.str.val = apc_xstrdup(fullpath, allocate);
+                        }
+                    }
+                }
+            }
+        }
+#endif 
     }
 
 #ifdef ZEND_ENGINE_2
-    if(flags->has_jumps) {
+    if(flags == NULL || flags->has_jumps) {
         apc_fixup_op_array_jumps(dst,src);
     }
 #endif
@@ -1485,14 +1562,18 @@ apc_class_t* apc_copy_new_classes(zend_op_array* op_array, int old_count, apc_ma
 static void my_destroy_zval_ptr(zval** src, apc_free_t deallocate)
 {
     assert(src != NULL);
-    my_destroy_zval(src[0], deallocate);
-    deallocate(src[0]);
+    if(my_destroy_zval(src[0], deallocate) == SUCCESS) {
+        deallocate(src[0]);
+    }
 }
 /* }}} */
 
 /* {{{ my_destroy_zval */
-static void my_destroy_zval(zval* src, apc_free_t deallocate)
+static int my_destroy_zval(zval* src, apc_free_t deallocate)
 {
+    zval **tmp;
+    TSRMLS_FETCH();
+
     switch (src->type & ~IS_CONSTANT_INDEX) {
     case IS_RESOURCE:
     case IS_BOOL:
@@ -1510,6 +1591,17 @@ static void my_destroy_zval(zval* src, apc_free_t deallocate)
         break;
     
     case IS_ARRAY:
+    
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        if(APCG(copied_zvals)) {
+            if(zend_hash_index_find(APCG(copied_zvals), (ulong)src, (void**)&tmp) == SUCCESS) {
+                (*tmp)->refcount--;
+                return FAILURE;
+            } 
+            zend_hash_index_update(APCG(copied_zvals), (ulong)src, (void**)&src, sizeof(zval*), NULL);
+        }
+        /* fall through */
+
     case IS_CONSTANT_ARRAY:
         my_free_hashtable(src->value.ht,
                           (ht_free_fun_t) my_free_zval_ptr,
@@ -1529,6 +1621,8 @@ static void my_destroy_zval(zval* src, apc_free_t deallocate)
     default:
         assert(0);
     }
+
+    return SUCCESS;
 }
 /* }}} */
 
@@ -1851,12 +1945,28 @@ void apc_free_classes(apc_class_t* src, apc_free_t deallocate)
 void apc_free_zval(zval* src, apc_free_t deallocate)
 {
     if (src != NULL) {
-        my_destroy_zval(src, deallocate);
-        deallocate(src);
+        if(my_destroy_zval(src, deallocate) == SUCCESS) {
+            deallocate(src);
+        }
     }
 }
 /* }}} */
 
+
+/* Used only by my_prepare_op_array_for_execution */
+#define APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION()                                                \
+                         /* The fetch is only required if auto_globals_jit=1  */                \
+                        if(zo->op2.u.EA.type == ZEND_FETCH_GLOBAL &&                            \
+                            zo->op1.op_type == IS_CONST &&                                      \
+                            zo->op1.u.constant.type == IS_STRING &&                             \
+                            zo->op1.u.constant.value.str.val[0] == '_') {                       \
+                                                                                                \
+                            znode* varname = &zo->op1;                                          \
+                            (void)zend_is_auto_global(varname->u.constant.value.str.val,        \
+                                                          varname->u.constant.value.str.len     \
+                                                          TSRMLS_CC);                           \
+                        }                                                                       \
+
 /* {{{ my_prepare_op_array_for_execution */
 static int my_prepare_op_array_for_execution(zend_op_array* dst, zend_op_array* src TSRMLS_DC) 
 {
@@ -1865,78 +1975,119 @@ static int my_prepare_op_array_for_execution(zend_op_array* dst, zend_op_array*
      *   - If the opcode stream contain mutable data, ensure a copy.
      *   - Fixup array jumps in the same loop.
      */
-    int i;
-    int needcopy = 0;
+    int i=src->last;
+    zend_op *zo;
+    zend_op *dzo;
 #ifdef ZEND_ENGINE_2
-    apc_opflags_t * flags = (apc_opflags_t*)&(src->reserved);
-#endif
+    apc_opflags_t * flags = APCG(reserved_offset) != -1 ? 
+                                (apc_opflags_t*) & (src->reserved[APCG(reserved_offset)]) : NULL;
+    int needcopy = flags ? flags->deep_copy : 1;
+    /* auto_globals_jit was not in php4 */
+    int do_prepare_fetch_global = PG(auto_globals_jit) && (flags == NULL || flags->unknown_global);
+
+#define FETCH_AUTOGLOBAL(member) do { \
+    if(flags && flags->member == 1) { \
+        zend_is_auto_global(#member,\
+                            (sizeof(#member) - 1)\
+                            TSRMLS_CC);\
+    } \
+}while(0); 
+            
+    FETCH_AUTOGLOBAL(_GET);
+    FETCH_AUTOGLOBAL(_POST);
+    FETCH_AUTOGLOBAL(_COOKIE);
+    FETCH_AUTOGLOBAL(_SERVER);
+    FETCH_AUTOGLOBAL(_ENV);
+    FETCH_AUTOGLOBAL(_FILES);
+    FETCH_AUTOGLOBAL(_REQUEST);
 
-#ifdef ZEND_ENGINE_2
-    needcopy = flags->deep_copy;
 #else
-    needcopy = 1;
-#endif
+    int needcopy = 0;
+    int do_prepare_fetch_global = 0;
+    int j = 0;
 
+    for(j = 0; j < src->last; j++) {
+        zo = &src->opcodes[j];
+        
+        if( ((zo->op1.op_type == IS_CONST &&
+              zo->op1.u.constant.type == IS_CONSTANT_ARRAY)) ||  
+            ((zo->op2.op_type == IS_CONST &&
+              zo->op2.u.constant.type == IS_CONSTANT_ARRAY))) {
+            needcopy = 1;
+        }
+    }
+#endif
+    
     if(needcopy) {
+
         dst->opcodes = (zend_op*) apc_xmemcpy(src->opcodes, 
                                     sizeof(zend_op) * src->last,
                                     apc_php_malloc);
-    }
+        zo = src->opcodes;
+        dzo = dst->opcodes;
+        while(i > 0) {
 
-    for (i = 0; i < src->last; i++) {
-        zend_op *zo = src->opcodes+i;
-        zend_op *dzo = dst->opcodes+i;
+            if( ((zo->op1.op_type == IS_CONST &&
+                  zo->op1.u.constant.type == IS_CONSTANT_ARRAY)) ||  
+                ((zo->op2.op_type == IS_CONST &&
+                  zo->op2.u.constant.type == IS_CONSTANT_ARRAY))) {
 
-        if(needcopy && 
-            ((zo->op1.op_type == IS_CONST &&
-              zo->op1.u.constant.type == IS_CONSTANT_ARRAY) ||
-             (zo->op1.op_type == IS_CONST &&
-              zo->op1.u.constant.type == IS_CONSTANT_ARRAY))) {
-
-            if(!(my_copy_zend_op(dzo, zo, apc_php_malloc, apc_php_free))) {
-                assert(0); /* emalloc failed or a bad constant array */
+                if(!(my_copy_zend_op(dzo, zo, apc_php_malloc, apc_php_free))) {
+                    assert(0); /* emalloc failed or a bad constant array */
+                }
             }
-        }
-        
-        switch(zo->opcode) {
+            
 #ifdef ZEND_ENGINE_2
-            case ZEND_JMP:
-                if(needcopy) {
+            switch(zo->opcode) {
+                case ZEND_JMP:
                     dzo->op1.u.jmp_addr = dst->opcodes + 
                                             (zo->op1.u.jmp_addr - src->opcodes);
-                }
-                break;
-            case ZEND_JMPZ:
-            case ZEND_JMPNZ:
-            case ZEND_JMPZ_EX:
-            case ZEND_JMPNZ_EX:
-                if(needcopy) {
+                    break;
+                case ZEND_JMPZ:
+                case ZEND_JMPNZ:
+                case ZEND_JMPZ_EX:
+                case ZEND_JMPNZ_EX:
                     dzo->op2.u.jmp_addr = dst->opcodes + 
                                             (zo->op2.u.jmp_addr - src->opcodes);
-                }
-                break;
-            /* auto_globals_jit was not in php4 */
-            case ZEND_FETCH_R:
-            case ZEND_FETCH_W:
-                if(PG(auto_globals_jit) && flags->use_globals)
-                {
-                     /* The fetch is only required if auto_globals_jit=1  */
-                    if(zo->op2.u.EA.type == ZEND_FETCH_GLOBAL &&
-                        zo->op1.op_type == IS_CONST && 
-                        zo->op1.u.constant.type == IS_STRING &&
-                        zo->op1.u.constant.value.str.val[0] == '_') {
-
-                        znode* varname = &zo->op1; 
-                        (void)zend_is_auto_global(varname->u.constant.value.str.val, 
-                                                      varname->u.constant.value.str.len 
-                                                      TSRMLS_CC);
+                    break;
+                case ZEND_FETCH_R:
+                case ZEND_FETCH_W:
+                case ZEND_FETCH_IS:
+                case ZEND_FETCH_FUNC_ARG:
+                    if(do_prepare_fetch_global)
+                    {
+                        APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION();
                     }
-                }
-                break;
+                    break;
+                default:
+                    break;
+            }
 #endif
-            default:
-                break;
+            i--;
+            zo++;
+            dzo++;
         }
+#ifdef ZEND_ENGINE_2
+    } else {  /* !needcopy */
+        /* The fetch is only required if auto_globals_jit=1  */
+        if(do_prepare_fetch_global)
+        {
+            zo = src->opcodes;
+            while(i > 0) {
+
+                if(zo->opcode == ZEND_FETCH_R || 
+                   zo->opcode == ZEND_FETCH_W ||
+                   zo->opcode == ZEND_FETCH_IS ||
+                   zo->opcode == ZEND_FETCH_FUNC_ARG 
+                  ) {
+                    APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION();
+                }
+
+                i--;
+                zo++;
+            }
+        }
+#endif
     }
     return 1;
 }
@@ -1945,9 +2096,6 @@ static int my_prepare_op_array_for_execution(zend_op_array* dst, zend_op_array*
 /* {{{ apc_copy_op_array_for_execution */
 zend_op_array* apc_copy_op_array_for_execution(zend_op_array* dst, zend_op_array* src TSRMLS_DC)
 {
-#ifdef ZEND_ENGINE_2
-    apc_opflags_t * flags = (apc_opflags_t*)&(src->reserved);
-#endif
     if(dst == NULL) {
         dst = (zend_op_array*) emalloc(sizeof(src[0]));
     }
@@ -1992,10 +2140,18 @@ zend_class_entry* apc_copy_class_entry_for_execution(zend_class_entry* src, int
     memcpy(dst, src, sizeof(src[0]));
 
 #ifdef ZEND_ENGINE_2
-    /* These are slots to be populated later by ADD_INTERFACE insns */
-    dst->interfaces = apc_php_malloc(sizeof(zend_class_entry*) * 
-                                        src->num_interfaces);
-    memset(dst->interfaces, 0, sizeof(zend_class_entry*) * src->num_interfaces);
+    if(src->num_interfaces)
+    {
+        /* These are slots to be populated later by ADD_INTERFACE insns */
+        dst->interfaces = apc_php_malloc(
+                            sizeof(zend_class_entry*) * src->num_interfaces);
+        memset(dst->interfaces, 0, 
+                            sizeof(zend_class_entry*) * src->num_interfaces);
+    }
+    else
+    {
+        /* assert(dst->interfaces == NULL); */
+    }
 #endif
 
 #ifndef ZEND_ENGINE_2    
@@ -2078,6 +2234,36 @@ zend_class_entry* apc_copy_class_entry_for_execution(zend_class_entry* src, int
 }
 /* }}} */
 
+/* {{{ apc_free_class_entry_after_execution */
+void apc_free_class_entry_after_execution(zend_class_entry* src)
+{
+#ifdef ZEND_ENGINE_2
+    if(src->num_interfaces > 0 && src->interfaces) {
+        apc_php_free(src->interfaces);
+        src->interfaces = NULL;
+        src->num_interfaces = 0;
+    }
+    /* my_destroy_hashtable() does not play nice with refcounts */
+
+    zend_hash_clean(&src->default_static_members);
+    if(src->static_members != &(src->default_static_members))
+    {
+        zend_hash_destroy(src->static_members);
+        apc_php_free(src->static_members);
+        src->static_members = NULL;
+    }
+    else
+    {
+        src->static_members = NULL;
+    }
+    zend_hash_clean(&src->default_properties);
+    zend_hash_clean(&src->constants_table);
+#endif
+
+    /* TODO: more cleanup */
+}
+/* }}} */
+
 #ifdef ZEND_ENGINE_2
 
 /* {{{ my_fixup_function */
@@ -2171,9 +2357,9 @@ static void my_fixup_hashtable(HashTable *ht, ht_fixup_fun_t fixup, zend_class_e
 static int my_check_copy_function(Bucket* p, va_list args)
 {
     zend_class_entry* src = va_arg(args, zend_class_entry*);
-    zend_class_entry* parent = src->parent;
     zend_function* zf = (zend_function*)p->pData;
 #ifndef ZEND_ENGINE_2
+    zend_class_entry* parent = src->parent;
     zend_function* parent_fn = NULL;
 #endif
 
@@ -2325,6 +2511,15 @@ static int my_check_copy_static_member(Bucket* p, va_list args)
 /* }}} */
 #endif
 
+/* {{{ apc_register_optimizer(apc_optimize_function_t optimizer)
+ *      register a optimizer callback function, returns the previous callback
+ */
+apc_optimize_function_t apc_register_optimizer(apc_optimize_function_t optimizer TSRMLS_DC) {
+    apc_optimize_function_t old_optimizer = APCG(apc_optimize_function);
+    APCG(apc_optimize_function) = optimizer;
+    return old_optimizer;
+}
+
 /*
  * Local variables:
  * tab-width: 4
