@@ -28,7 +28,7 @@
 
  */
 
-/* $Id: apc_main.c,v 3.83 2006/09/04 23:56:39 rasmus Exp $ */
+/* $Id: apc_main.c,v 3.97 2007/03/22 16:03:59 gopalv Exp $ */
 
 #include "apc_php.h"
 #include "apc_main.h"
@@ -80,7 +80,7 @@ static int install_function(apc_function_t fn TSRMLS_DC)
                       NULL);
 
     if (status == FAILURE) {
-        /* zend_error(E_ERROR, "Cannot redeclare %s()", fn.name); */
+        /* apc_eprint("Cannot redeclare %s()", fn.name); */
     }
 
     return status;
@@ -97,7 +97,6 @@ static int install_class(apc_class_t cl TSRMLS_DC)
     zend_class_entry** allocated_ce = NULL;
 #endif
 
-#ifdef ZEND_ENGINE_2    
 
     /* Special case for mangled names. Mangled names are unique to a file.
      * There is no way two classes with the same mangled name will occur,
@@ -113,6 +112,7 @@ static int install_class(apc_class_t cl TSRMLS_DC)
         }
     }
     
+#ifdef ZEND_ENGINE_2    
     /*
      * XXX: We need to free this somewhere...
      */
@@ -154,7 +154,7 @@ static int install_class(apc_class_t cl TSRMLS_DC)
 #endif
         if (status == FAILURE) {
             if(APCG(report_autofilter)) {
-                zend_error(E_WARNING, "Dynamic inheritance detected for class %s", cl.name);
+                apc_wprint("Dynamic inheritance detected for class %s", cl.name);
             }
             class_entry->parent = NULL;
             return status;
@@ -164,7 +164,11 @@ static int install_class(apc_class_t cl TSRMLS_DC)
             parent = *parent_ptr;
 #endif 
             class_entry->parent = parent;
+#ifdef ZEND_ENGINE_2
             zend_do_inheritance(class_entry, parent TSRMLS_CC);
+#else
+            zend_do_inheritance(class_entry, parent);
+#endif
         }
 
 
@@ -187,7 +191,7 @@ static int install_class(apc_class_t cl TSRMLS_DC)
 #endif                           
 
     if (status == FAILURE) {
-        zend_error(E_ERROR, "Cannot redeclare class %s", cl.name);
+        apc_eprint("Cannot redeclare class %s", cl.name);
     } 
     return status;
 }
@@ -208,7 +212,7 @@ static int uninstall_class(apc_class_t cl TSRMLS_DC)
                            cl.name_len+1);
 #endif                           
     if (status == FAILURE) {
-        zend_error(E_ERROR, "Cannot delete class %s", cl.name);
+        apc_eprint("Cannot delete class %s", cl.name);
     } 
     return status;
 }
@@ -256,14 +260,16 @@ default_compile:
 
     cache_entry->autofiltered = 1;
     if(APCG(report_autofilter)) {
-        zend_error(E_WARNING, "Autofiltering %s", h->opened_path);
+        apc_wprint("Autofiltering %s", h->opened_path);
     }
 
     if(cache_entry->data.file.classes) {
         for(ii = 0; ii < i ; ii++) {
-            uninstall_class(cache_entry->data.file.classes[i] TSRMLS_CC);
+            uninstall_class(cache_entry->data.file.classes[ii] TSRMLS_CC);
         }
     }
+    
+    apc_stack_pop(APCG(cache_stack)); /* pop out cache_entry */
 
     /* cannot free up cache data yet, it maybe in use */
     
@@ -291,12 +297,12 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
     char *path;
     size_t mem_size;
 
-    if (!APCG(enabled) || apc_cache_busy(apc_cache)) {
+    if (!APCG(enabled) || (apc_cache_busy(apc_cache) && !APCG(localcache))) {
 		return old_compile_file(h, type TSRMLS_CC);
 	}
 
     /* check our regular expression filters */
-    if (apc_compiled_filters) {
+    if (APCG(filters) && apc_compiled_filters) {
         int ret = apc_regex_match_array(apc_compiled_filters, h->filename);
         if(ret == APC_NEGATIVE_MATCH || (ret != APC_POSITIVE_MATCH && !APCG(cache_by_default))) {
             return old_compile_file(h, type TSRMLS_CC);
@@ -323,9 +329,15 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
     if (!apc_cache_make_file_key(&key, h->filename, PG(include_path), t TSRMLS_CC)) {
         return old_compile_file(h, type TSRMLS_CC);
     }
-    
-    /* search for the file in the cache */
-    cache_entry = apc_cache_find(apc_cache, key, t);
+
+    if(APCG(localcache)) {
+        /* search for the file in the local cache */
+        cache_entry = apc_local_cache_find(APCG(lcache), key, t);
+    } else {
+        /* search for the file in the cache */
+        cache_entry = apc_cache_find(apc_cache, key, t);
+    }
+
     if (cache_entry != NULL && !cache_entry->autofiltered) {
         int dummy = 1;
         if (h->opened_path == NULL) {
@@ -357,6 +369,10 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
         return old_compile_file(h, type TSRMLS_CC);
     }
     
+    if(apc_cache_busy(apc_cache) && APCG(localcache)) {
+        /* possibly local cache returned NULL because cache is busy */
+		return old_compile_file(h, type TSRMLS_CC);
+	}
 
     /* remember how many functions and classes existed before compilation */
     num_functions = zend_hash_num_elements(CG(function_table));
@@ -464,7 +480,6 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
     }
     APCG(mem_size_ptr) = NULL;
     cache_entry->mem_size = mem_size;
-    cache_entry->autofiltered = 0;
 
     if ((ret = apc_cache_insert(apc_cache, key, cache_entry, t)) != 1) {
         apc_cache_free_entry(cache_entry);
@@ -497,6 +512,7 @@ int apc_module_init(int module_number TSRMLS_DC)
 #endif
     apc_cache = apc_cache_create(APCG(num_files_hint), APCG(gc_ttl), APCG(ttl));
     apc_user_cache = apc_cache_create(APCG(user_entries_hint), APCG(gc_ttl), APCG(user_ttl));
+
     apc_compiled_filters = apc_regex_compile_array(APCG(filters));
 
     /* override compilation */
@@ -556,12 +572,35 @@ int apc_module_shutdown(TSRMLS_D)
 
 /* }}} */
 
+/* {{{ process init and shutdown */
+int apc_process_init(int module_number TSRMLS_DC)
+{
+    int minttl = (APCG(gc_ttl) >  APCG(ttl) ? APCG(ttl) : APCG(gc_ttl))/2;
+    int size = APCG(localcache_size);
+    if(APCG(initialized) && APCG(localcache)) {
+        /* TTL is 2 mins by default */
+        APCG(lcache) = apc_local_cache_create(apc_cache, size, minttl ? minttl : 120); 
+    }
+    return 0;
+}
+
+int apc_process_shutdown(TSRMLS_D)
+{
+    if(APCG(initialized) && APCG(localcache) && APCG(lcache)) {
+        apc_local_cache_destroy(APCG(lcache));
+        APCG(lcache) = NULL;
+    }
+	return 0;
+}
+/* }}} */
+
 /* {{{ request init and shutdown */
 
 int apc_request_init(TSRMLS_D)
 {
     apc_stack_clear(APCG(cache_stack));
     APCG(slam_rand) = -1;
+    APCG(copied_zvals) = NULL;
     return 0;
 }
 
@@ -580,15 +619,55 @@ void apc_deactivate(TSRMLS_D)
      * the reference counts on active cache entries in `my_execute`.
      */
     while (apc_stack_size(APCG(cache_stack)) > 0) {
+        int i;
+        zend_class_entry* zce = NULL;
+        void ** centry = (void*)(&zce);
+#ifdef ZEND_ENGINE_2
+        zend_class_entry** pzce = NULL;
+#endif
+        
         apc_cache_entry_t* cache_entry =
             (apc_cache_entry_t*) apc_stack_pop(APCG(cache_stack));
+
+        if (cache_entry->data.file.functions) {
+            for (i = 0; cache_entry->data.file.functions[i].function != NULL; i++) {
+                zend_hash_del(EG(function_table),
+                    cache_entry->data.file.functions[i].name,
+                    cache_entry->data.file.functions[i].name_len+1);
+            }
+        }
+
+        if (cache_entry->data.file.classes) {
+            for (i = 0; cache_entry->data.file.classes[i].class_entry != NULL; i++) {
+#ifdef ZEND_ENGINE_2
+                centry = (void**)&pzce; /* a triple indirection to get zend_class_entry*** */
+#endif
+                if(zend_hash_find(EG(class_table), 
+                    cache_entry->data.file.classes[i].name,
+                    cache_entry->data.file.classes[i].name_len+1,
+                    (void**)centry) == FAILURE)
+                {
+                    /* double inclusion of conditional classes ends up failing 
+                     * this lookup the second time around.
+                     */
+                    continue;
+                }
+
+#ifdef ZEND_ENGINE_2
+                zce = *pzce;
+#endif
+                zend_hash_del(EG(class_table),
+                    cache_entry->data.file.classes[i].name,
+                    cache_entry->data.file.classes[i].name_len+1);
+                
+                apc_free_class_entry_after_execution(zce);
+            }
+        }
         apc_cache_release(apc_cache, cache_entry);
     }
-    /* Safety net */
-#if 0
-    apc_sma_unlock();
-    apc_cache_unlock(apc_cache);
-#endif
+    if(APCG(localcache)) {
+        apc_local_cache_cleanup(APCG(lcache)); 
+    }
 }
 /* }}} */
 
