@@ -26,7 +26,7 @@
 
  */
 
-/* $Id: apc_sma.c,v 1.53 2006/06/09 23:41:22 rasmus Exp $ */
+/* $Id: apc_sma.c,v 1.63 2007/04/02 09:02:57 gopalv Exp $ */
 
 #include "apc_sma.h"
 #include "apc.h"
@@ -53,14 +53,13 @@ static size_t sma_segsize;          /* size of each shm segment */
 static int* sma_segments;           /* array of shm segment ids */
 static void** sma_shmaddrs;         /* array of shm segment addresses */
 static int sma_lastseg = 0;         /* index of MRU segment */
-static int sma_lock;                /* sempahore to serialize access */
 
 typedef struct header_t header_t;
 struct header_t {
-    size_t segsize;    /* size of entire segment */
-    size_t avail;      /* bytes available (not necessarily contiguous) */
-    size_t nfoffset;   /* start next fit search from this offset       */
-    size_t padding;    /* Unused - just padding up to an even boundary */
+    apc_lck_t sma_lock;     /* segment lock, MUST BE ALIGNED for futex locks */
+    size_t segsize;         /* size of entire segment */
+    size_t avail;           /* bytes available (not necessarily contiguous) */
+    size_t nfoffset;        /* start next fit search from this offset       */
 #if ALLOC_DISTRIBUTION
     size_t adist[30];
 #endif
@@ -78,12 +77,16 @@ struct header_t {
 static volatile size_t block_id = 0;
 #endif
 
+#define APC_SMA_CANARIES 1   
+
 typedef struct block_t block_t;
 struct block_t {
     size_t size;       /* size of this block */
     size_t next;       /* offset in segment of next free block */
-#ifdef __APC_SMA_DEBUG__
+#ifdef APC_SMA_CANARIES
     size_t canary;     /* canary to check for memory overwrites */
+#endif
+#ifdef __APC_SMA_DEBUG__
     size_t id;         /* identifier for the memory block */ 
 #endif
 };
@@ -95,17 +98,27 @@ struct block_t {
 #define BLOCKAT(offset) ((block_t*)((char *)shmaddr + offset))
 #define OFFSET(block) ((size_t)(((char*)block) - (char*)shmaddr))
 
+/* Canary macros for setting, checking and resetting memory canaries */
+#ifdef APC_SMA_CANARIES
+    #define SET_CANARY(v) (v)->canary = 0x42424242
+    #define CHECK_CANARY(v) assert((v)->canary == 0x42424242)
+    #define RESET_CANARY(v) (v)->canary = -42
+#else
+    #define SET_CANARY(v) 
+    #define CHECK_CANARY(v)
+    #define RESET_CANARY(v)
+#endif
+
+
 #ifdef max
 #undef max
 #endif
 #define max(a, b) ((a) > (b) ? (a) : (b))
 
-/* {{{ alignword: returns x, aligned to the system's word boundary */
-static int alignword(int x)
-{
-    typedef union { void* p; int i; long l; double d; void (*f)(); } word_t;
-    return sizeof(word_t) * (1 + ((x-1)/sizeof(word_t)));
-}
+/* {{{ ALIGNWORD: pad up x, aligned to the system's word boundary */
+typedef union { void* p; int i; long l; double d; void (*f)(); } apc_word_t;
+#define ALIGNWORD(x) (sizeof(apc_word_t) * (1 + (((x)-1)/sizeof(apc_word_t))))
+#define MINBLOCKSIZE (ALIGNWORD(1) + ALIGNWORD(sizeof(block_t)))
 /* }}} */
 
 /* {{{ sma_allocate: tries to allocate size bytes in a segment */
@@ -118,8 +131,9 @@ static int sma_allocate(void* shmaddr, size_t size)
     size_t realsize;        /* actual size of block needed, including header */
     size_t last_offset;     /* save the last search offset */
     int wrapped=0;
+    const size_t block_size = ALIGNWORD(sizeof(struct block_t));
 
-    realsize = alignword(size + alignword(sizeof(struct block_t)));
+    realsize = ALIGNWORD(size + block_size);
 
     /*
      * First, insure that the segment contains at least realsize free bytes,
@@ -139,10 +153,16 @@ static int sma_allocate(void* shmaddr, size_t size)
     } else {    
         prv = BLOCKAT(sizeof(header_t));
     }
+   
+    CHECK_CANARY(prv);
+
     while (prv->next != header->nfoffset) {
         cur = BLOCKAT(prv->next);
-        /* If it fits perfectly or it fits after a split, stop searching */
-        if (cur->size == realsize || (cur->size > (sizeof(block_t) + realsize))) {
+#ifdef __APC_SMA_DEBUG__
+        CHECK_CANARY(cur);
+#endif
+        /* If it can fit realiszie bytes in cur block, stop searching */
+        if (cur->size >= realsize) {
             prvnextfit = prv;
             break;
         }
@@ -153,6 +173,9 @@ static int sma_allocate(void* shmaddr, size_t size)
         /* Check to see if we need to wrap around and search from the top */
         if(header->nfoffset && prv->next == 0) {
             prv = BLOCKAT(sizeof(header_t));
+#ifdef __APC_SMA_DEBUG__
+            CHECK_CANARY(prv);
+#endif
             last_offset = 0;
             wrapped = 1;
         } 
@@ -166,14 +189,11 @@ static int sma_allocate(void* shmaddr, size_t size)
     prv = prvnextfit;
     cur = BLOCKAT(prv->next);
 
-    /* update the block header */
-    header->avail -= realsize;
-#if ALLOC_DISTRIBUTION
-    header->adist[(int)(log(size)/log(2))]++;
-#endif
+    CHECK_CANARY(prv);
+    CHECK_CANARY(cur);
 
-    if (cur->size == realsize) {
-        /* cur is a perfect fit for realsize; just unlink it */
+    if (cur->size == realsize || (cur->size > realsize && cur->size < (realsize + (MINBLOCKSIZE * 2)))) {
+        /* cur is big enough for realsize, but too small to split - unlink it */
         prv->next = cur->next;
     }
     else {
@@ -189,16 +209,27 @@ static int sma_allocate(void* shmaddr, size_t size)
         nxt = BLOCKAT(prv->next);
         nxt->next = nxtoffset;  /* Re-link the shortened block */
         nxt->size = oldsize - realsize;  /* and fix the size */
+        SET_CANARY(nxt);
+#ifdef __APC_SMA_DEBUG__
+        nxt->id = -1;
+#endif
     }
+
+    /* update the block header */
+    header->avail -= cur->size;
+#if ALLOC_DISTRIBUTION
+    header->adist[(int)(log(size)/log(2))]++;
+#endif
+
     header->nfoffset = last_offset;
 
+    SET_CANARY(cur);
 #ifdef __APC_SMA_DEBUG__
     cur->id = ++block_id;
-    cur->canary = 0x42424242;
     fprintf(stderr, "allocate(realsize=%d,size=%d,id=%d)\n", (int)(size), (int)(cur->size), cur->id);
 #endif
 
-    return OFFSET(cur) + alignword(sizeof(struct block_t));
+    return OFFSET(cur) + block_size;
 }
 /* }}} */
 
@@ -209,25 +240,37 @@ static int sma_deallocate(void* shmaddr, int offset)
     block_t* cur;       /* the new block to insert */
     block_t* prv;       /* the block before cur */
     block_t* nxt;       /* the block after cur */
-    size_t size;           /* size of deallocated block */
+    size_t size;        /* size of deallocated block */
 
-    offset -= alignword(sizeof(struct block_t));
+    offset -= ALIGNWORD(sizeof(struct block_t));
     assert(offset >= 0);
 
     /* find position of new block in free list */
+    cur = BLOCKAT(offset);
     prv = BLOCKAT(sizeof(header_t));
+   
+    CHECK_CANARY(cur);
+
+#ifdef __APC_SMA_DEBUG__
+    CHECK_CANARY(prv);
+    fprintf(stderr, "free(%p, size=%d,id=%d)\n", cur, (int)(cur->size), cur->id);
+#endif
     while (prv->next != 0 && prv->next < offset) {
         prv = BLOCKAT(prv->next);
+#ifdef __APC_SMA_DEBUG__
+        CHECK_CANARY(prv);
+#endif
     }
+    
+    CHECK_CANARY(prv);
 
     /* insert new block after prv */
-    cur = BLOCKAT(offset);
     cur->next = prv->next;
     prv->next = offset;
 
 #ifdef __APC_SMA_DEBUG__
-    fprintf(stderr, "free(size=%d,id=%d)\n", (int)(cur->size), cur->id);
-    assert(cur->canary == 0x42424242);
+    CHECK_CANARY(cur);
+    cur->id = -1;
 #endif
     
     /* update the block header */
@@ -239,14 +282,21 @@ static int sma_deallocate(void* shmaddr, int offset)
         /* cur and prv share an edge, combine them */
         prv->size += cur->size;
         prv->next = cur->next;
+        RESET_CANARY(cur);
         cur = prv;
     }
 
     nxt = BLOCKAT(cur->next);
+
     if (((char *)cur) + cur->size == (char *) nxt) {
         /* cur and nxt shared an edge, combine them */
         cur->size += nxt->size;
         cur->next = nxt->next;
+#ifdef __APC_SMA_DEBUG__
+        CHECK_CANARY(nxt);
+        nxt->id = -1; /* assert this or set it ? */
+#endif
+        RESET_CANARY(nxt);
     }
     header->nfoffset = 0;  /* Reset the next fit search marker */
 
@@ -286,8 +336,6 @@ void apc_sma_init(int numseg, int segsize, char *mmap_file_mask)
     sma_segments = (int*) apc_emalloc(sma_numseg*sizeof(int));
     sma_shmaddrs = (void**) apc_emalloc(sma_numseg*sizeof(void*));
     
-    sma_lock = apc_lck_create(NULL, 0, 1);
-
     for (i = 0; i < sma_numseg; i++) {
         header_t*   header;
         block_t*    block;
@@ -304,9 +352,10 @@ void apc_sma_init(int numseg, int segsize, char *mmap_file_mask)
         shmaddr = sma_shmaddrs[i];
     
         header = (header_t*) shmaddr;
+        apc_lck_create(NULL, 0, 1, header->sma_lock);
         header->segsize = sma_segsize;
         header->avail = sma_segsize - sizeof(header_t) - sizeof(block_t) -
-                        alignword(sizeof(int));
+                        ALIGNWORD(sizeof(int));
         header->nfoffset = 0;
 #if ALLOC_DISTRIBUTION
        	{
@@ -317,9 +366,17 @@ void apc_sma_init(int numseg, int segsize, char *mmap_file_mask)
         block = BLOCKAT(sizeof(header_t));
         block->size = 0;
         block->next = sizeof(header_t) + sizeof(block_t);
+        SET_CANARY(block);
+#ifdef __APC_SMA_DEBUG__
+        block->id = -1;
+#endif
         block = BLOCKAT(block->next);
         block->size = header->avail;
         block->next = 0;
+        SET_CANARY(block);
+#ifdef __APC_SMA_DEBUG__
+        block->id = -1;
+#endif
     }
 }
 /* }}} */
@@ -332,13 +389,13 @@ void apc_sma_cleanup()
     assert(sma_initialized);
 
     for (i = 0; i < sma_numseg; i++) {
+        apc_lck_destroy(((header_t*)sma_shmaddrs[i])->sma_lock);
 #if APC_MMAP
         apc_unmap(sma_shmaddrs[i], sma_segments[i]);
 #else
         apc_shm_detach(sma_shmaddrs[i]);
 #endif
     }
-    apc_lck_destroy(sma_lock);
     sma_initialized = 0;
     apc_efree(sma_segments);
     apc_efree(sma_shmaddrs);
@@ -353,17 +410,19 @@ void* apc_sma_malloc(size_t n)
 
     TSRMLS_FETCH();
     assert(sma_initialized);
-    LOCK(sma_lock);
+    LOCK(((header_t*)sma_shmaddrs[sma_lastseg])->sma_lock);
 
     off = sma_allocate(sma_shmaddrs[sma_lastseg], n);
     if (off != -1) {
         void* p = (void *)(((char *)(sma_shmaddrs[sma_lastseg])) + off);
         if (APCG(mem_size_ptr) != NULL) { *(APCG(mem_size_ptr)) += n; }
-        UNLOCK(sma_lock);
+        UNLOCK(((header_t*)sma_shmaddrs[sma_lastseg])->sma_lock);
         return p;
     }
+    UNLOCK(((header_t*)sma_shmaddrs[sma_lastseg])->sma_lock);
 
     for (i = 0; i < sma_numseg; i++) {
+        LOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
         if (i == sma_lastseg) {
             continue;
         }
@@ -371,13 +430,13 @@ void* apc_sma_malloc(size_t n)
         if (off != -1) {
             void* p = (void *)(((char *)(sma_shmaddrs[i])) + off);
             if (APCG(mem_size_ptr) != NULL) { *(APCG(mem_size_ptr)) += n; }
-            UNLOCK(sma_lock);
+            UNLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
             sma_lastseg = i;
             return p;
         }
+        UNLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
     }
 
-    UNLOCK(sma_lock);
     return NULL;
 }
 /* }}} */
@@ -410,6 +469,8 @@ char* apc_sma_strdup(const char* s)
 void apc_sma_free(void* p)
 {
     int i;
+    size_t offset;
+    size_t d_size;
     TSRMLS_FETCH();
 
     if (p == NULL) {
@@ -417,55 +478,61 @@ void apc_sma_free(void* p)
     }
 
     assert(sma_initialized);
-    LOCK(sma_lock);
 
     for (i = 0; i < sma_numseg; i++) {
-        size_t d_size = (size_t)((char *)p - (char *)(sma_shmaddrs[i]));
-        if (p >= sma_shmaddrs[i] && d_size < sma_segsize) {
-            sma_deallocate(sma_shmaddrs[i], d_size);
+        LOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
+        offset = (size_t)((char *)p - (char *)(sma_shmaddrs[i]));
+        if (p >= sma_shmaddrs[i] && offset < sma_segsize) {
+            d_size = sma_deallocate(sma_shmaddrs[i], offset);
             if (APCG(mem_size_ptr) != NULL) { *(APCG(mem_size_ptr)) -= d_size; }
-            UNLOCK(sma_lock);
+            UNLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
             return;
         }
+        UNLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
     }
 
     apc_eprint("apc_sma_free: could not locate address %p", p);
-    UNLOCK(sma_lock);
 }
 /* }}} */
 
 /* {{{ apc_sma_info */
-apc_sma_info_t* apc_sma_info()
+apc_sma_info_t* apc_sma_info(zend_bool limited)
 {
     apc_sma_info_t* info;
     apc_sma_link_t** link;
     int i;
-
+	char* shmaddr;
+	block_t* prv;
+	
     if (!sma_initialized) {
         return NULL;
     }
 
     info = (apc_sma_info_t*) apc_emalloc(sizeof(apc_sma_info_t));
     info->num_seg = sma_numseg;
-    info->seg_size = sma_segsize - sizeof(header_t) - sizeof(block_t) - alignword(sizeof(int));
+    info->seg_size = sma_segsize - sizeof(header_t) - sizeof(block_t) - ALIGNWORD(sizeof(int));
 
     info->list = apc_emalloc(info->num_seg * sizeof(apc_sma_link_t*));
     for (i = 0; i < sma_numseg; i++) {
         info->list[i] = NULL;
     }
 
-    RDLOCK(sma_lock);
+    if(limited) return info;
 
     /* For each segment */
     for (i = 0; i < sma_numseg; i++) {
-        char* shmaddr = sma_shmaddrs[i];
-        block_t* prv = BLOCKAT(sizeof(header_t));
+        RDLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
+        shmaddr = sma_shmaddrs[i];
+        prv = BLOCKAT(sizeof(header_t));
 
         link = &info->list[i];
 
         /* For each block in this segment */
         while (prv->next != 0) {
             block_t* cur = BLOCKAT(prv->next);
+#ifdef __APC_SMA_DEBUG__
+            CHECK_CANARY(cur);
+#endif
 
             *link = apc_emalloc(sizeof(apc_sma_link_t));
             (*link)->size = cur->size;
@@ -475,9 +542,9 @@ apc_sma_info_t* apc_sma_info()
 
             prv = cur;
         }
+        UNLOCK(((header_t*)sma_shmaddrs[i])->sma_lock);
     }
 
-    UNLOCK(sma_lock);
     return info;
 }
 /* }}} */
@@ -520,12 +587,6 @@ size_t *apc_sma_get_alloc_distribution(void) {
     return header->adist; 
 }
 #endif
-/* {{{ apc_sma_unlock */
-void apc_sma_unlock()
-{
-    UNLOCK(sma_lock);
-}
-/* }}} */
 
 #if 0
 /* {{{ apc_sma_check_integrity */
