@@ -26,9 +26,8 @@
 
  */
 
-/* $Id: php_apc.c,v 3.104 2006/09/01 21:59:16 pollita Exp $ */
+/* $Id: php_apc.c,v 3.140 2007/03/28 07:14:54 gopalv Exp $ */
 
-#include "php_apc.h"
 #include "apc_zend.h"
 #include "apc_cache.h"
 #include "apc_main.h"
@@ -38,6 +37,8 @@
 #include "php_ini.h"
 #include "ext/standard/info.h"
 #include "SAPI.h"
+#include "rfc1867.h"
+#include "php_apc.h"
 #if PHP_API_VERSION <= 20020918
 #if HAVE_APACHE
 #ifdef APC_PHP4_STAT
@@ -53,8 +54,11 @@ PHP_FUNCTION(apc_clear_cache);
 PHP_FUNCTION(apc_sma_info);
 PHP_FUNCTION(apc_store);
 PHP_FUNCTION(apc_fetch);
+PHP_FUNCTION(apc_delete);
+PHP_FUNCTION(apc_compile_file);
 PHP_FUNCTION(apc_define_constants);
 PHP_FUNCTION(apc_load_constants);
+PHP_FUNCTION(apc_add);
 /* }}} */
 
 /* {{{ ZEND_DECLARE_MODULE_GLOBALS(apc) */
@@ -74,8 +78,20 @@ static void php_apc_init_globals(zend_apc_globals* apc_globals TSRMLS_DC)
     apc_globals->slam_defense = 0;
     apc_globals->mem_size_ptr = NULL;
     apc_globals->fpstat = 1;
-    apc_globals->write_lock = 0;
+    apc_globals->stat_ctime = 0;
+    apc_globals->write_lock = 1;
     apc_globals->report_autofilter = 0;
+    apc_globals->apc_optimize_function = NULL;
+#ifdef MULTIPART_EVENT_FORMDATA
+    apc_globals->rfc1867 = 0;
+#endif
+    apc_globals->copied_zvals = NULL;
+#ifdef ZEND_ENGINE_2
+    apc_globals->reserved_offset = -1;
+#endif
+    apc_globals->localcache = 0;
+    apc_globals->localcache_size = 0;
+    apc_globals->lcache = NULL;
 }
 
 static void php_apc_shutdown_globals(zend_apc_globals* apc_globals TSRMLS_DC)
@@ -102,39 +118,60 @@ static void php_apc_shutdown_globals(zend_apc_globals* apc_globals TSRMLS_DC)
 
 /* {{{ PHP_INI */
 
-static PHP_INI_MH(OnUpdate_filters)
+static PHP_INI_MH(OnUpdate_filters) /* {{{ */
 {
     APCG(filters) = apc_tokenize(new_value, ',');
     return SUCCESS;
 }
+/* }}} */
+
+static PHP_INI_MH(OnUpdateShmSegments) /* {{{ */
+{
+#if APC_MMAP
+    if(atoi(new_value)!=1) {
+        php_error_docref(NULL TSRMLS_CC, E_WARNING, "apc.shm_segments setting ignored in MMAP mode");
+    }
+    APCG(shm_segments) = 1; 
+#else
+    APCG(shm_segments) = atoi(new_value);
+#endif
+    return SUCCESS;
+}
+/* }}} */
+
 
 #ifdef ZEND_ENGINE_2
 #define OnUpdateInt OnUpdateLong
 #endif
 
 PHP_INI_BEGIN()
-STD_PHP_INI_BOOLEAN("apc.enabled",      "1",    PHP_INI_ALL, OnUpdateBool,              enabled,         zend_apc_globals, apc_globals)
-STD_PHP_INI_ENTRY("apc.shm_segments",   "1",    PHP_INI_SYSTEM, OnUpdateInt,            shm_segments,    zend_apc_globals, apc_globals)
+STD_PHP_INI_BOOLEAN("apc.enabled",      "1",    PHP_INI_SYSTEM, OnUpdateBool,              enabled,         zend_apc_globals, apc_globals)
+STD_PHP_INI_ENTRY("apc.shm_segments",   "1",    PHP_INI_SYSTEM, OnUpdateShmSegments,       shm_segments,    zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.shm_size",       "30",   PHP_INI_SYSTEM, OnUpdateInt,            shm_size,        zend_apc_globals, apc_globals)
-STD_PHP_INI_BOOLEAN("apc.optimization", "0",    PHP_INI_ALL, OnUpdateInt,               optimization,    zend_apc_globals, apc_globals)
 STD_PHP_INI_BOOLEAN("apc.include_once_override", "0", PHP_INI_SYSTEM, OnUpdateBool,     include_once,    zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.num_files_hint", "1000", PHP_INI_SYSTEM, OnUpdateInt,            num_files_hint,  zend_apc_globals, apc_globals)
-STD_PHP_INI_ENTRY("apc.user_entries_hint", "100", PHP_INI_SYSTEM, OnUpdateInt,          user_entries_hint, zend_apc_globals, apc_globals)
+STD_PHP_INI_ENTRY("apc.user_entries_hint", "4096", PHP_INI_SYSTEM, OnUpdateInt,          user_entries_hint, zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.gc_ttl",         "3600", PHP_INI_SYSTEM, OnUpdateInt,            gc_ttl,           zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.ttl",            "0",    PHP_INI_SYSTEM, OnUpdateInt,            ttl,              zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.user_ttl",       "0",    PHP_INI_SYSTEM, OnUpdateInt,            user_ttl,         zend_apc_globals, apc_globals)
 #if APC_MMAP
 STD_PHP_INI_ENTRY("apc.mmap_file_mask",  NULL,  PHP_INI_SYSTEM, OnUpdateString,         mmap_file_mask,   zend_apc_globals, apc_globals)
 #endif
-    PHP_INI_ENTRY("apc.filters",        NULL,     PHP_INI_SYSTEM, OnUpdate_filters)
-STD_PHP_INI_BOOLEAN("apc.cache_by_default", "1",  PHP_INI_SYSTEM, OnUpdateBool,         cache_by_default, zend_apc_globals, apc_globals)
+PHP_INI_ENTRY("apc.filters",        NULL,     PHP_INI_SYSTEM, OnUpdate_filters)
+STD_PHP_INI_BOOLEAN("apc.cache_by_default", "1",  PHP_INI_ALL, OnUpdateBool,         cache_by_default, zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.slam_defense", "0",      PHP_INI_SYSTEM, OnUpdateInt,            slam_defense,     zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.file_update_protection", "2", PHP_INI_SYSTEM, OnUpdateInt,file_update_protection,  zend_apc_globals, apc_globals)
 STD_PHP_INI_BOOLEAN("apc.enable_cli", "0",      PHP_INI_SYSTEM, OnUpdateBool,           enable_cli,       zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.max_file_size", "1M",    PHP_INI_SYSTEM, OnUpdateInt,            max_file_size,    zend_apc_globals, apc_globals)
 STD_PHP_INI_BOOLEAN("apc.stat", "1",            PHP_INI_SYSTEM, OnUpdateBool,           fpstat,           zend_apc_globals, apc_globals)
+STD_PHP_INI_BOOLEAN("apc.stat_ctime", "0",      PHP_INI_SYSTEM, OnUpdateBool,           stat_ctime,       zend_apc_globals, apc_globals)
 STD_PHP_INI_BOOLEAN("apc.write_lock", "1",      PHP_INI_SYSTEM, OnUpdateBool,           write_lock,       zend_apc_globals, apc_globals)
 STD_PHP_INI_BOOLEAN("apc.report_autofilter", "0", PHP_INI_SYSTEM, OnUpdateBool,         report_autofilter,zend_apc_globals, apc_globals)
+#ifdef MULTIPART_EVENT_FORMDATA
+STD_PHP_INI_BOOLEAN("apc.rfc1867", "0", PHP_INI_SYSTEM, OnUpdateBool, rfc1867, zend_apc_globals, apc_globals)
+#endif
+STD_PHP_INI_BOOLEAN("apc.localcache", "0", PHP_INI_SYSTEM, OnUpdateBool, localcache, zend_apc_globals, apc_globals)
+STD_PHP_INI_ENTRY("apc.localcache.size", "512", PHP_INI_SYSTEM, OnUpdateInt, localcache_size,  zend_apc_globals, apc_globals)
 PHP_INI_END()
 
 /* }}} */
@@ -151,13 +188,28 @@ static PHP_MINFO_FUNCTION(apc)
 #else
     php_info_print_table_row(2, "MMAP Support", "Disabled");
 #endif
-    php_info_print_table_row(2, "Revision", "$Revision: 3.104 $");
+#if APC_SEM_LOCKS
+    php_info_print_table_row(2, "Locking type", "IPC Semaphore");
+#elif APC_FUTEX_LOCKS
+    php_info_print_table_row(2, "Locking type", "Linux Futex Locks");
+#elif APC_PTHREADMUTEX_LOCKS
+    php_info_print_table_row(2, "Locking type", "pthread mutex Locks");
+#elif APC_SPIN_LOCKS
+    php_info_print_table_row(2, "Locking type", "spin Locks");
+#else
+    php_info_print_table_row(2, "Locking type", "File Locks");
+#endif
+    php_info_print_table_row(2, "Revision", "$Revision: 3.140 $");
     php_info_print_table_row(2, "Build Date", __DATE__ " " __TIME__);
     php_info_print_table_end();
     DISPLAY_INI_ENTRIES();
 }
 /* }}} */
 
+#ifdef MULTIPART_EVENT_FORMDATA
+extern int apc_rfc1867_progress(unsigned int event, void *event_data, void **extra TSRMLS_DC);
+#endif
+
 /* {{{ PHP_MINIT_FUNCTION(apc) */
 static PHP_MINIT_FUNCTION(apc)
 {
@@ -167,12 +219,23 @@ static PHP_MINIT_FUNCTION(apc)
 
     /* Disable APC in cli mode unless overridden by apc.enable_cli */
     if(!APCG(enable_cli) && !strcmp(sapi_module.name, "cli")) {
-	APCG(enabled) = 0;
+        APCG(enabled) = 0;
     }
 
     if (APCG(enabled)) {
-        apc_module_init(module_number TSRMLS_CC);
-        apc_zend_init(TSRMLS_C);
+        if(APCG(initialized)) {
+            apc_process_init(module_number TSRMLS_CC);
+        } else {
+            apc_module_init(module_number TSRMLS_CC);
+            apc_zend_init(TSRMLS_C);
+            apc_process_init(module_number TSRMLS_CC);
+#ifdef MULTIPART_EVENT_FORMDATA
+            /* File upload progress tracking */
+            if(APCG(rfc1867)) {
+                php_rfc1867_callback = apc_rfc1867_progress;
+            }
+#endif
+        }
     }
 
     return SUCCESS;
@@ -183,14 +246,16 @@ static PHP_MINIT_FUNCTION(apc)
 static PHP_MSHUTDOWN_FUNCTION(apc)
 {
     if(APCG(enabled)) {
+        apc_process_shutdown(TSRMLS_C);
         apc_zend_shutdown(TSRMLS_C);
         apc_module_shutdown(TSRMLS_C);
-#ifdef ZTS
-        ts_free_id(apc_globals_id);
-#else
+#ifndef ZTS
         php_apc_shutdown_globals(&apc_globals);
 #endif
     }
+#ifdef ZTS
+    ts_free_id(apc_globals_id);
+#endif
     UNREGISTER_INI_ENTRIES();
     return SUCCESS;
 }
@@ -252,6 +317,32 @@ PHP_FUNCTION(apc_cache_info)
     add_assoc_long(return_value, "expunges", info->expunges);
     add_assoc_long(return_value, "mem_size", info->mem_size);
     add_assoc_long(return_value, "num_entries", info->num_entries);
+    add_assoc_long(return_value, "num_inserts", info->num_inserts);
+#ifdef MULTIPART_EVENT_FORMDATA
+    add_assoc_long(return_value, "file_upload_progress", 1);
+#else
+    add_assoc_long(return_value, "file_upload_progress", 0);
+#endif
+#if APC_MMAP
+    add_assoc_stringl(return_value, "memory_type", "mmap", sizeof("mmap")-1, 1);
+#else
+    add_assoc_stringl(return_value, "memory_type", "IPC shared", sizeof("IPC shared")-1, 1);
+#endif
+#if APC_SEM_LOCKS
+    add_assoc_stringl(return_value, "locking_type", "IPC semaphore", sizeof("IPC semaphore")-1, 1);
+#elif APC_FUTEX_LOCKS
+    add_assoc_stringl(return_value, "locking_type", "Linux Futex", sizeof("Linux Futex")-1, 1);
+#elif APC_PTHREADMUTEX_LOCKS
+    add_assoc_stringl(return_value, "locking_type", "pthread mutex", sizeof("pthread mutex")-1, 1);
+#elif APC_SPIN_LOCKS
+    add_assoc_stringl(return_value, "locking_type", "spin", sizeof("spin")-1, 1);
+#else
+    add_assoc_stringl(return_value, "locking_type", "file", sizeof("file")-1, 1);
+#endif
+    if(limited) {
+        apc_cache_free_info(info);
+        return;
+    }
     
     ALLOC_INIT_ZVAL(list);
     array_init(list);
@@ -337,18 +428,19 @@ PHP_FUNCTION(apc_clear_cache)
 }
 /* }}} */
 
-/* {{{ proto array apc_sma_info() */
+/* {{{ proto array apc_sma_info([bool limited]) */
 PHP_FUNCTION(apc_sma_info)
 {
     apc_sma_info_t* info;
     zval* block_lists;
     int i;
+    zend_bool limited = 0;
 
-    if (ZEND_NUM_ARGS() != 0) {
-        WRONG_PARAM_COUNT;
+    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|b", &limited) == FAILURE) {
+        return;
     }
 
-    info = apc_sma_info();
+    info = apc_sma_info(limited);
 
     if(!info) {
         php_error_docref(NULL TSRMLS_CC, E_WARNING, "No APC SMA info available.  Perhaps APC is disabled via apc.enabled?");
@@ -359,6 +451,12 @@ PHP_FUNCTION(apc_sma_info)
     add_assoc_long(return_value, "num_seg", info->num_seg);
     add_assoc_long(return_value, "seg_size", info->seg_size);
     add_assoc_long(return_value, "avail_mem", apc_sma_get_avail_mem());
+
+    if(limited) {
+        apc_sma_free_info(info);
+        return;
+    }
+
 #if ALLOC_DISTRIBUTION
     {
         size_t *adist = apc_sma_get_alloc_distribution();
@@ -399,7 +497,7 @@ PHP_FUNCTION(apc_sma_info)
 /* }}} */
 
 /* {{{ _apc_store */
-static int _apc_store(char *strkey, int strkey_len, const zval *val, const unsigned int ttl TSRMLS_DC) {
+int _apc_store(char *strkey, int strkey_len, const zval *val, const unsigned int ttl, const int exclusive TSRMLS_DC) {
     apc_cache_entry_t *entry;
     apc_cache_key_t key;
     time_t t;
@@ -437,8 +535,7 @@ static int _apc_store(char *strkey, int strkey_len, const zval *val, const unsig
         return 0;
     }
 
-
-    if (!apc_cache_user_insert(apc_user_cache, key, entry, t TSRMLS_CC)) {
+    if (!apc_cache_user_insert(apc_user_cache, key, entry, t, exclusive TSRMLS_CC)) {
         APCG(mem_size_ptr) = NULL;
         apc_cache_free_entry(entry);
         apc_cache_expunge(apc_cache,t);
@@ -469,7 +566,26 @@ PHP_FUNCTION(apc_store) {
 
     if(!strkey_len) RETURN_FALSE;
 
-    if(_apc_store(strkey, strkey_len, val, (unsigned int)ttl TSRMLS_CC)) RETURN_TRUE;
+    if(_apc_store(strkey, strkey_len, val, (unsigned int)ttl, 0 TSRMLS_CC)) RETURN_TRUE;
+    RETURN_FALSE;
+}
+/* }}} */
+
+/* {{{ proto int apc_add(string key, zval var [, ttl ])
+ */
+PHP_FUNCTION(apc_add) {
+    zval *val;
+    char *strkey;
+    int strkey_len;
+    long ttl = 0L;
+
+    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz|l", &strkey, &strkey_len, &val, &ttl) == FAILURE) {
+        return;
+    }
+
+    if(!strkey_len) RETURN_FALSE;
+
+    if(_apc_store(strkey, strkey_len, val, (unsigned int)ttl, 1 TSRMLS_CC)) RETURN_TRUE;
     RETURN_FALSE;
 }
 /* }}} */
@@ -478,9 +594,39 @@ void *apc_erealloc_wrapper(void *ptr, size_t size) {
     return _erealloc(ptr, size, 0 ZEND_FILE_LINE_CC ZEND_FILE_LINE_EMPTY_CC);
 }
 
-/* {{{ proto mixed apc_fetch(string key)
+/* {{{ RETURN_ZVAL for php4 */
+#if !defined(ZEND_ENGINE_2) && !defined(RETURN_ZVAL)
+#define RETURN_ZVAL(zv, copy, dtor) { RETVAL_ZVAL(zv, copy, dtor); return; } 
+#define RETVAL_ZVAL(zv, copy, dtor)     ZVAL_ZVAL(return_value, zv, copy, dtor)
+#define ZVAL_ZVAL(z, zv, copy, dtor) {  \
+        int is_ref, refcount;           \
+        is_ref = (z)->is_ref;           \
+        refcount = (z)->refcount;       \
+        *(z) = *(zv);                   \
+        if (copy) {                     \
+            zval_copy_ctor(z);          \
+        }                               \
+        if (dtor) {                     \
+            if (!copy) {                \
+                ZVAL_NULL(zv);          \
+            }                           \
+            zval_ptr_dtor(&zv);         \
+        }                               \
+        (z)->is_ref = is_ref;           \
+        (z)->refcount = refcount;       \
+    }
+#endif
+/* }}} */
+
+/* {{{ proto mixed apc_fetch(mixed key)
  */
 PHP_FUNCTION(apc_fetch) {
+    zval *key;
+    HashTable *hash;
+    HashPosition hpos;
+    zval **hentry;
+    zval *result;
+    zval *result_entry;
     char *strkey;
     int strkey_len;
     apc_cache_entry_t* entry;
@@ -488,12 +634,10 @@ PHP_FUNCTION(apc_fetch) {
 
     if(!APCG(enabled)) RETURN_FALSE;
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &strkey, &strkey_len) == FAILURE) {
+    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &key) == FAILURE) {
         return;
     }
 
-    if(!strkey_len) RETURN_FALSE;
-
 #if PHP_API_VERSION <= 20041225
 #if HAVE_APACHE && defined(APC_PHP4_STAT)
     t = ((request_rec *)SG(server_context))->request_time;
@@ -504,14 +648,49 @@ PHP_FUNCTION(apc_fetch) {
     t = sapi_get_request_time(TSRMLS_C);
 #endif
 
-    entry = apc_cache_user_find(apc_user_cache, strkey, strkey_len + 1, t);
-    if(entry) {
-        /* deep-copy returned shm zval to emalloc'ed return_value */
-        apc_cache_fetch_zval(return_value, entry->data.user.val, apc_php_malloc, apc_php_free);
-        apc_cache_release(apc_user_cache, entry);
+    if(Z_TYPE_P(key) != IS_STRING && Z_TYPE_P(key) != IS_ARRAY) {
+        convert_to_string(key);
+    }
+    
+    if(Z_TYPE_P(key) == IS_STRING) {
+        strkey = Z_STRVAL_P(key);
+        strkey_len = Z_STRLEN_P(key);
+        if(!strkey_len) RETURN_FALSE;
+        entry = apc_cache_user_find(apc_user_cache, strkey, strkey_len + 1, t);
+        if(entry) {
+            /* deep-copy returned shm zval to emalloc'ed return_value */
+            apc_cache_fetch_zval(return_value, entry->data.user.val, apc_php_malloc, apc_php_free);
+            apc_cache_release(apc_user_cache, entry);
+        } else {
+            RETURN_FALSE;
+        }
+    } else if(Z_TYPE_P(key) == IS_ARRAY) {
+        hash = Z_ARRVAL_P(key);
+        MAKE_STD_ZVAL(result);
+        array_init(result); 
+        zend_hash_internal_pointer_reset_ex(hash, &hpos);
+        while(zend_hash_get_current_data_ex(hash, (void**)&hentry, &hpos) == SUCCESS) {
+            if(Z_TYPE_PP(hentry) != IS_STRING) {
+                apc_wprint("apc_fetch() expects a string or array of strings.");
+                RETURN_FALSE;
+            }
+            entry = apc_cache_user_find(apc_user_cache, Z_STRVAL_PP(hentry), Z_STRLEN_PP(hentry) + 1, t);
+            if(entry) {
+                /* deep-copy returned shm zval to emalloc'ed return_value */
+                MAKE_STD_ZVAL(result_entry);
+                apc_cache_fetch_zval(result_entry, entry->data.user.val, apc_php_malloc, apc_php_free);
+                apc_cache_release(apc_user_cache, entry);
+                zend_hash_add(Z_ARRVAL_P(result), Z_STRVAL_PP(hentry), Z_STRLEN_PP(hentry) +1, &result_entry, sizeof(zval*), NULL);
+            } /* don't set values we didn't find */
+            zend_hash_move_forward_ex(hash, &hpos);
+        }
+        RETURN_ZVAL(result, 0, 1);
     } else {
+        apc_wprint("apc_fetch() expects a string or array of strings.");
         RETURN_FALSE;
     }
+
+    return;
 }
 /* }}} */
 
@@ -571,6 +750,9 @@ static void _apc_define_constants(zval *constants, zend_bool case_sensitive TSRM
         c.flags = case_sensitive;
         c.name = zend_strndup(const_key, const_key_len);
         c.name_len = const_key_len;
+#ifdef ZEND_ENGINE_2
+        c.module_number = PHP_USER_CONSTANT;
+#endif
         zend_register_constant(&c TSRMLS_CC);
 
         zend_hash_move_forward_ex(Z_ARRVAL_P(constants), &pos);
@@ -593,7 +775,7 @@ PHP_FUNCTION(apc_define_constants) {
     if(!strkey_len) RETURN_FALSE;
 
     _apc_define_constants(constants, case_sensitive TSRMLS_CC);
-    if(_apc_store(strkey, strkey_len, constants, 0 TSRMLS_CC)) RETURN_TRUE;
+    if(_apc_store(strkey, strkey_len, constants, 0, 0 TSRMLS_CC)) RETURN_TRUE;
     RETURN_FALSE;
 } /* }}} */
 
@@ -635,6 +817,101 @@ PHP_FUNCTION(apc_load_constants) {
 }
 /* }}} */
 
+/* {{{ proto boolean apc_compile_file(string filename)
+ */
+PHP_FUNCTION(apc_compile_file) {
+    char *filename;
+    int filename_len;
+    zend_file_handle file_handle;
+    zend_op_array *op_array;
+    long slam_defense = 0;
+    char** filters = NULL;
+    zend_bool cache_by_default = 1;
+    HashTable cg_function_table, cg_class_table, eg_function_table, eg_class_table;
+    HashTable *cg_orig_function_table, *cg_orig_class_table, *eg_orig_function_table, *eg_orig_class_table;
+
+    if(!APCG(enabled)) RETURN_FALSE;
+
+    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &filename, &filename_len) == FAILURE) {
+        return;
+    }
+
+    if(!filename) RETURN_FALSE;
+
+    /* If slam defense is active, temporarily disable */
+    if(APCG(slam_defense)) {
+        slam_defense = APCG(slam_defense);
+        APCG(slam_defense) = 0;
+    }
+   
+    /* If filter is active, temporarily disable */
+    if(APCG(filters) != NULL) {
+        filters = APCG(filters);
+        APCG(filters) = NULL;
+    }
+
+    /* If cache_by_default is off, temporarily enable */
+    if(!APCG(cache_by_default)) {
+        cache_by_default = APCG(cache_by_default);
+        APCG(cache_by_default) = 1;
+    }
+
+    /* Replace function/class tables to avoid namespace conflicts */
+    zend_hash_init_ex(&cg_function_table, 100, NULL, ZEND_FUNCTION_DTOR, 1, 0);
+    cg_orig_function_table = CG(function_table);
+    CG(function_table) = &cg_function_table;
+    zend_hash_init_ex(&cg_class_table, 10, NULL, ZEND_CLASS_DTOR, 1, 0);
+    cg_orig_class_table = CG(class_table);
+    CG(class_table) = &cg_class_table;
+    zend_hash_init_ex(&eg_function_table, 100, NULL, ZEND_FUNCTION_DTOR, 1, 0);
+    eg_orig_function_table = EG(function_table);
+    EG(function_table) = &eg_function_table;
+    zend_hash_init_ex(&eg_class_table, 10, NULL, ZEND_CLASS_DTOR, 1, 0);
+    eg_orig_class_table = EG(class_table);
+    EG(class_table) = &eg_class_table;
+    
+    /* Compile the file, loading it into the cache */
+    file_handle.type = ZEND_HANDLE_FILENAME;
+    file_handle.filename = filename;
+    file_handle.free_filename = 0;
+    file_handle.opened_path = NULL;
+    zend_try {
+        op_array = zend_compile_file(&file_handle, ZEND_INCLUDE TSRMLS_CC);
+    } zend_catch {
+        apc_eprint("Error compiling %s in apc_compile_file.", filename);
+        op_array = NULL;
+    } zend_end_try();
+
+    /* Return class/function tables to previous states, destroy temp tables */
+    CG(function_table) = cg_orig_function_table;
+    zend_hash_destroy(&cg_function_table);
+    CG(class_table) = cg_orig_class_table;
+    zend_hash_destroy(&cg_class_table);
+    EG(function_table) = eg_orig_function_table;
+    zend_hash_destroy(&eg_function_table);
+    EG(class_table) = eg_orig_class_table;
+    zend_hash_destroy(&eg_class_table);
+    
+    /* Restore global settings */
+    APCG(slam_defense) = slam_defense;
+    APCG(filters) = filters;
+    APCG(cache_by_default) = cache_by_default;
+
+    if(op_array == NULL) { RETURN_FALSE; }
+
+    /* Free up everything */
+    zend_destroy_file_handle(&file_handle TSRMLS_CC);
+#ifdef ZEND_ENGINE_2
+    destroy_op_array(op_array TSRMLS_CC);
+#else
+    destroy_op_array(op_array);
+#endif
+    efree(op_array);
+
+    RETURN_TRUE;
+}
+/* }}} */
+
 /* {{{ apc_functions[] */
 function_entry apc_functions[] = {
 	PHP_FE(apc_cache_info,          NULL)
@@ -645,6 +922,7 @@ function_entry apc_functions[] = {
 	PHP_FE(apc_delete,              NULL)
 	PHP_FE(apc_define_constants,    NULL)
 	PHP_FE(apc_load_constants,      NULL)
+	PHP_FE(apc_compile_file,        NULL)
 	{NULL, 		NULL,				NULL}
 };
 /* }}} */
